import pytorch_lightning as pl

class ZoobotEncoder(pl.LightningModule):
    # very simple wrapper to turn pytorch model into lightning module
    # useful when we want to use lightning to make predictions with our encoder
    # (i.e. to get representations)

    def __init__(self, encoder, pyramid=False) -> None:
        super().__init__()
        self.encoder = encoder  # plain pytorch module e.g. Sequential
        if pyramid:
            raise NotImplementedError('Will eventually support resetting timm classifier to get FPN features')

    def forward(self, x):
        return self.encoder(x)



import logging
import time
import datetime
from typing import List

import pandas as pd
import torch
import pytorch_lightning as pl

from zoobot.shared import save_predictions
from galaxy_datasets.pytorch.galaxy_datamodule import GalaxyDataModule


def predict(catalog: pd.DataFrame, model: pl.LightningModule, n_samples: int, label_cols: List, save_loc: str, datamodule_kwargs={}, trainer_kwargs={}):
    """
    Use trained model to make predictions on a catalog of galaxies.

    Args:
        catalog (pd.DataFrame): catalog of galaxies to make predictions on. Must include `file_loc` and `id_str` columns.
        model (pl.LightningModule): with which to make predictions. Probably ZoobotTree, FinetuneableZoobotClassifier, FinetuneableZoobotTree, or ZoobotEncoder.
        n_samples (int): num. of forward passes to make per galaxy. Useful to marginalise over augmentations/test-time dropout.
        label_cols (List): Names for prediction columns. Only for your convenience - has no effect on predictions.
        save_loc (str): desired name of file recording the predictions
        datamodule_kwargs (dict, optional): Passed to GalaxyDataModule. Use to e.g. add custom augmentations. Defaults to {}.
        trainer_kwargs (dict, optional): Passed to pl.Trainer. Defaults to {}.
    """

    image_id_strs = list(catalog['id_str'].astype(str))

    predict_datamodule = GalaxyDataModule(
        label_cols=None,  # not using label_cols to load labels, we're only using it to name our predictions
        predict_catalog=catalog,  # no need to specify the other catalogs
        # will use the default transforms unless overridden with datamodule_kwargs
        # 
        **datamodule_kwargs  # e.g. batch_size, resize_size, crop_scale_bounds, etc.
    )
    # with this stage arg, will only use predict_catalog 
    # crucial to specify the stage, or will error (as missing other catalogs)
    predict_datamodule.setup(stage='predict')  
    # for images in predict_datamodule.predict_dataloader():
    #     print(images)
    #     print(images.shape)


    # set up trainer (again)
    trainer = pl.Trainer(
        max_epochs=-1,  # does nothing in this context, suppresses warning
        inference_mode=True,  # no grads needed
        **trainer_kwargs  # e.g. gpus
    )

    # from here, very similar to tensorflow version - could potentially refactor

    logging.info('Beginning predictions')
    start = datetime.datetime.fromtimestamp(time.time())
    logging.info('Starting at: {}'.format(start.strftime('%Y-%m-%d %H:%M:%S')))

    logging.info(len(trainer.predict(model, predict_datamodule)))

    # trainer.predict gives list of tensors, each tensor being predictions for a batch. Concat on axis 0.
    # range(n_samples) list comprehension repeats this, for dropout-permuted predictions. Stack to create new last axis.
    # final shape (n_galaxies, n_answers, n_samples)
    predictions = torch.stack([torch.concat(trainer.predict(model, predict_datamodule), dim=0) for n in range(n_samples)], dim=-1).numpy()
    logging.info('Predictions complete - {}'.format(predictions.shape))

    logging.info(f'Saving predictions to {save_loc}')

    if save_loc.endswith('.csv'):      # save as pandas df
        save_predictions.predictions_to_csv(predictions, image_id_strs, label_cols, save_loc)
    elif save_loc.endswith('.hdf5'):
        save_predictions.predictions_to_hdf5(predictions, image_id_strs, label_cols, save_loc)
    else:
        logging.warning('Save format of {} not recognised - assuming csv'.format(save_loc))
        save_predictions.predictions_to_csv(predictions, image_id_strs, label_cols, save_loc)

    logging.info(f'Predictions saved to {save_loc}')

    end = datetime.datetime.fromtimestamp(time.time())
    logging.info('Completed at: {}'.format(end.strftime('%Y-%m-%d %H:%M:%S')))
    logging.info('Time elapsed: {}'.format(end - start))



from typing import List

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from PIL import Image
import tensorflow as tf
import tensorflow_probability as tfp

# continuous posteriors, looks much prettier but technically not defined for non-integer votes

def show_binary_predictions(samples_list: List, observed_votes: np.ndarray, image_paths: List, xlabel: str, answer_index=0, n_examples=5):
    """
    Visualise dirichlet distribution posteriors against images and true observed_votes.
    Shows each model (element in sample_list) as solid colour
    Shows each dropout pass from that model as trace in that colour
    Shows observed observed_votes in black dashed.

    Args:
        samples_list (List): where each element is the predicted concentrations from a model and is shaped like (galaxy, answer, dropout_pass). Should be 2 answers i.e. dim1=2. One model is fine, one dropout pass is fine.
        observed_votes (np.ndarray): actual recorded responses, by galaxy. By tfp convention, needs to have both answers recorded explicitly e.g. [[2, 8], [4, 6]] not [2, 4]. This implies the total votes, which is useful.
        image_paths (list): paths to images, will imshow next to posteriors. Align with samples_list
        xlabel (str]): for final graph. Plotting only, no other effect.
        answer_index (int, optional): Which answer index to display on graph. Defaults to 0. Setting 1 will flip the posteriors.
        n_examples (int, optional): Number of galaxies (rows) to show. Defaults to 5.

    Returns:
        fig, axes: matplotlib figure, axes of posteriors and galaxy images row-by-row
    """
    
    cycler = mpl.rcParams['axes.prop_cycle']
    # https://matplotlib.org/cycler/
    colors = [c['color'] for c in cycler]
    assert len(colors) > len(samples_list)

    fig, axes = plt.subplots(nrows=n_examples, ncols=2, figsize=(6.8, 3 * n_examples), gridspec_kw={'hspace': .4})

    for galaxy_index in range(n_examples):
        
        row = axes[galaxy_index]
        total_votes = observed_votes[galaxy_index].sum()
        actual_votes = observed_votes[galaxy_index][0]  # show first index by default

        votes = np.linspace(0., total_votes)
        x = np.stack([votes, total_votes-votes], axis=-1)  # also need the counts for other answer, no

        n_dropout_samples = samples_list[0].shape[2]

        for model_n, samples in enumerate(samples_list):
            ax = row[answer_index]

            all_log_probs = []
            
            color = colors[model_n]

            for dropout_index in range(n_dropout_samples):
                concentrations = tf.constant(samples[galaxy_index, :, dropout_index].astype(np.float32))
                log_probs = tfp.distributions.DirichletMultinomial(total_votes, concentrations).prob(x)
                all_log_probs.append(log_probs)
                ax.plot(votes, log_probs, alpha=.15, color=color)
            all_log_probs = np.array(all_log_probs).mean(axis=0)
            ax.plot(votes, all_log_probs, linewidth=2., color=color)
            
            # avoid disappearing off the side if zero votes
            viz_offset = 0.004
            if actual_votes == 0:
                actual_votes += viz_offset * total_votes

            ax.axvline(actual_votes, color='k', linestyle='--', linewidth=2.)
            ax.set_xlim([0., total_votes])
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            ax.set_ylabel(r'$p$(votes)')
            
            ax = row[1]
            ax.imshow(np.array(Image.open(image_paths[galaxy_index])))
            ax.axis('off')
            
    # last row, left side
    ax = axes[-1][0]
    ax.set_xlabel(xlabel)

    return fig, axes


# if __name__ == '__main__':

    # example predictions to test with
    # from zoobot.predictions import predict_on_dataset

    # predictions = np.random.rand(100, 2)
    # image_paths = ['some_path_' + str(x) for x in np.arange(100)]
    # label_cols = ['ring', 'not_ring']
    # save_loc = '/Users/walml/repos/zoobot_private/test.hdf5'
    # predict_on_dataset.predictions_to_hdf5(predictions, image_paths, label_cols, save_loc)

    # version='decals'
    # label_cols = label_metadata.decals_label_cols
    # questions = label_metadata.decals_questions
    # schema = losses.Schema(label_cols, questions, version=version)

    # # question = schema.get_question('smooth-or-featured')
    # # question = schema.get_question('disk-edge-on')
    # question = schema.get_question('has-spiral-arms')
    # # question = schema.get_question('bar')
    # answer = question.answers[0]

    # catalog_loc = 'data/decals/decals_master_catalog.csv'
    # catalog = pd.read_csv(catalog_loc, dtype={'subject_id': str})  # original catalog
    # # catalog['file_loc'] = catalog['local_png_loc'].apply(lambda x: '/media/walml/beta/gz2' + x[32:])
    # catalog['file_loc'] = catalog['local_png_loc'].apply(lambda x: '/media/walml/beta1/decals' + x.replace('/data/phys-zooniverse/chri5177', ''))
    # retired = catalog[catalog['smooth-or-featured_total-votes'] > 35]


    # show_predictions(prediction_dfs, question, answer, n_examples=5, single_model=False)



# import pytest

# # cannot be run on mac m1 as tensorflow cannot be installed
# import tensorflow_probability as tfp
# import torch
# import numpy as np

# from zoobot.tensorflow.training import losses as tf_losses
# from zoobot.pytorch.training import losses as torch_losses


# @pytest.fixture
# def single_question_context():
#     # use factory pattern to get different values each time called within a test
#     # https://docs.pytest.org/en/6.2.x/fixture.html#factories-as-fixtures
#     def _single_question_context(batch_size=10, answers=3):
#         total_count = np.ones(batch_size) * 5
#         concentrations = np.random.rand(batch_size, answers)
#         # generated with tfp, but just to get some mock labels - doesn't bias
#         labels = tfp.distributions.DirichletMultinomial(total_count, concentrations).sample().numpy()
#         return labels, total_count, concentrations
#     return _single_question_context


# def test_single_question_loss(single_question_context):

#     labels, total_count, concentrations = single_question_context()

#     # tf will broadcast total_count e.g. 5 -> [5, 5, 5...] by batch shape
#     # tf will also accept int or float total_counts
#     log_prob_tf = tf_losses.get_dirichlet_neg_log_prob(labels, total_count, concentrations).numpy()

#     # pyro won't broadcast total count, needs to be = batch shape
#     # pyro will only accept total_count dtype = labels dtype
#     log_prob_torch = torch_losses.get_dirichlet_neg_log_prob(torch.from_numpy(labels), torch.Tensor(total_count), torch.from_numpy(concentrations)).numpy()

#     assert np.isclose(log_prob_tf, log_prob_torch).all()


# def test_multiquestion_loss(single_question_context):
#     labels_a, _, concentrations_a = single_question_context(answers=2)
#     labels_b, _, concentrations_b = single_question_context(answers=3)

#     # assert not np.isclose(labels_a, labels_b).all()  # will be diffferent data due to factory fixture

#     labels_both = np.concatenate([labels_a, labels_b], axis=1)
#     # total_count added up per question within func
#     concentrations_both = np.concatenate([concentrations_a, concentrations_b], axis=1)

#     question_index_groups = [(0, 1), (2, 4)]  # indices 0,1 for q a, 2,3,4 for q b

#     # tf accepts np args
#     log_prob_tf = tf_losses.calculate_multiquestion_loss(labels_both, concentrations_both, question_index_groups).numpy()
#     log_prob_torch = torch_losses.calculate_multiquestion_loss(torch.from_numpy(labels_both), torch.from_numpy(concentrations_both), question_index_groups).numpy()

#     assert np.isclose(log_prob_tf, log_prob_torch).all()



import tensorflow as tf

from zoobot.tensorflow.training import losses


class LossPerQuestion(tf.keras.metrics.Metric):

  def __init__(self, question_index_groups, name='loss_per_question'):
    super(LossPerQuestion, self).__init__(name=name)

    self.question_index_groups = question_index_groups

    self.multiq_loss_func = losses.get_multiquestion_loss(question_index_groups=self.question_index_groups, reduction=tf.keras.losses.Reduction.NONE)  # doesn't sum over questions    

    # dict of weights, keyed by question_n
    self.question_weights = {}
    for question_n in range(len(self.question_index_groups)):
        self.question_weights[question_n] = self.add_weight(name=f'questions/question_{question_n}_loss', initializer='zeros')

    # also track num. galaxies
    self.num_galaxies = self.add_weight(name='num_galaxies', initializer='zeros')

  def update_state(self, y_true, y_pred, sample_weight=None):

    multiq_loss = tf.cast(self.multiq_loss_func(y_true, y_pred), tf.float32)

    for question_n in range(len(self.question_index_groups)):
        total_loss_for_question_in_batch = tf.reduce_sum(multiq_loss[:, question_n])
        self.question_weights[question_n].assign_add(total_loss_for_question_in_batch)

    self.num_galaxies.assign_add(tf.cast(len(y_true), dtype=tf.float32))
 

  def result(self):

    metric_result = {}
    for weight in self.question_weights.values():
      # tf.print(weight/self.num_galaxies)
      # .ref() is the hashable string that you'd imagine .name would give, .name is some unhashable weird TF object 
      metric_result[weight.name] = weight/self.num_galaxies  # total loss for q across all batches, divide by total num galaxies

    # tf.print(metric_result)
    # return weight/self.num_galaxies
    return metric_result
    # return {'something': self.question_weights[0], 'something_else': self.num_galaxies}
    # TODO rename with 
    # return self.question_weights[0]



import logging
import os
from typing import Tuple

import pytorch_lightning as pl
from pytorch_lightning.strategies.ddp import DDPStrategy
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.callbacks.early_stopping import EarlyStopping

from galaxy_datasets.pytorch.galaxy_datamodule import GalaxyDataModule

from zoobot.pytorch.estimators import define_model


def train_default_zoobot_from_scratch(    
    # absolutely crucial arguments
    save_dir: str,  # save model here
    schema,  # answer these questions
    # input data - specify *either* catalog (to be split) or the splits themselves
    catalog=None,
    train_catalog=None,
    val_catalog=None,
    test_catalog=None,
    # training time parameters
    epochs=1000,
    patience=8,
    # model hparams
    architecture_name='efficientnet_b0',  # recently changed
    batch_size=128,
    dropout_rate=0.2,
    drop_connect_rate=0.2,
    learning_rate=1e-3,
    betas=(0.9, 0.999),
    weight_decay=0.01,
    scheduler_params={},
    # data and augmentation parameters
    color=False,
    resize_after_crop=224,
    crop_scale_bounds=(0.7, 0.8),
    crop_ratio_bounds=(0.9, 1.1),
    # hardware parameters
    nodes=1,
    gpus=2,
    num_workers=4,
    prefetch_factor=4,
    mixed_precision=False,
    # checkpointing / logging
    wandb_logger=None,
    checkpoint_file_template=None,
    auto_insert_metric_name=True,
    save_top_k=3,
    extra_callbacks=None,
    # replication parameters
    random_state=42
) -> Tuple[define_model.ZoobotTree, pl.Trainer]:
    """
    Train Zoobot from scratch on a big galaxy catalog.
    Zoobot is a base deep learning model (anything from timm, typically a CNN) plus a dirichlet head.
    Images are augmented using the default transforms (flips, rotations, zooms)
    from `the galaxy-datasets repo <https://github.com/mwalmsley/galaxy-datasets/blob/main/galaxy_datasets/transforms.py>`_.

    Once trained, Zoobot can be finetuned to new data.
    For finetuning, see zoobot/pytorch/training/finetune.py.
    Many pretrained models are already available - see :ref:`datanotes`.

    Args:
        save_dir (str): folder to save training logs and trained model checkpoints
        catalog (pd.DataFrame, optional): Galaxy catalog with columns `id_str` and `file_loc`. Will be automatically split to train and val (no test). Defaults to None. 
        train_catalog (pd.DataFrame, optional): As above, but already split by you for training. Defaults to None.
        val_catalog (pd.DataFrame, optional): As above, for validation. Defaults to None.
        test_catalog (pd.DataFrame, optional): As above, for testing. Defaults to None.
        epochs (int, optional): Max. number of epochs to train for. Defaults to 1000.
        patience (int, optional): Max. number of epochs to wait for any loss improvement before ending training. Defaults to 8.
        architecture_name (str, optional): Architecture to use. Passed to timm. Must be in timm.list_models(). Defaults to 'efficientnet_b0'.
        dropout_rate (float, optional): Randomly drop activations prior to the output layer, with this probability. Defaults to 0.2.
        drop_connect_rate (float, optional): Randomly drop blocks with this probability, for regularisation. For supported timm models only. Defaults to 0.2.
        learning_rate (float, optional): Base learning rate for AdamW. Defaults to 1e-3.
        betas (tuple, optional): Beta args (i.e. momentum) for adamW. Defaults to (0.9, 0.999).
        weight_decay (float, optional): Weight decay arg (i.e. L2 penalty) for AdamW. Defaults to 0.01.
        scheduler_params (dict, optional): Specify a learning rate scheduler. See code. Not recommended with AdamW. Defaults to {}.
        color (bool, optional): Train on RGB images rather than channel-averaged. Defaults to False.
        resize_after_crop (int, optional): Input image size. After all transforms, images will be resized to this size. Defaults to 224.
        crop_scale_bounds (tuple, optional): Off-center crop fraction (<1 means zoom in). Defaults to (0.7, 0.8).
        crop_ratio_bounds (tuple, optional): Aspect ratio of crop above. Defaults to (0.9, 1.1).
        nodes (int, optional): Multi-node training Unlikely to work on your cluster without tinkering. Defaults to 1 (i.e. one node).
        gpus (int, optional): Multi-GPU training. Uses distributed data parallel - essentially, full dataset is split by GPU. See torch docs. Defaults to 2.
        num_workers (int, optional): Processes for loading data. See torch dataloader docs. Should be < num cpus. Defaults to 4.
        prefetch_factor (int, optional): Num. batches to queue in memory per dataloader. See torch dataloader docs. Defaults to 4.
        mixed_precision (bool, optional): Use (mostly) half-precision to halve memory requirements. May cause instability. See Lightning Trainer docs. Defaults to False.
        wandb_logger (pl.loggers.wandb.WandbLogger, optional): Logger to track experiments on Weights and Biases. Defaults to None.
        checkpoint_file_template (str, optional): formatting for checkpoint filename. See Lightning docs. Defaults to None.
        auto_insert_metric_name (bool, optional): escape "/" in metric names when naming checkpoints. See Lightning docs. Defaults to True.
        save_top_k (int, optional): Keep the k best checkpoints. See Lightning docs. Defaults to 3.
        random_state (int, optional): Seed. Defaults to 42.

    Returns:
        Tuple[define_model.ZoobotTree, pl.Trainer]: Trained ZoobotTree model, and Trainer with which it was trained.
    """

    # some optional logging.debug calls recording cluster environment
    slurm_debugging_logs()

    # redundant when already called before this, but no harm doing twice
    pl.seed_everything(random_state)

    assert save_dir is not None
    if not os.path.isdir(save_dir):
        os.mkdir(save_dir)

    if color:
        logging.warning(
            'Training on color images, not converting to greyscale')
        channels = 3
    else:
        logging.info('Converting images to greyscale before training')
        channels = 1

    strategy = None
    plugins = None
    if (gpus is not None) and (gpus > 1):
        strategy = DDPStrategy(find_unused_parameters=False)  # static_graph=True TODO
        logging.info('Using multi-gpu training')
        if nodes > 1:  # I assume nobody is doing multi-node cpu training...
            logging.info('Using multi-node training')  # purely for your info
            # this is only needed for multi-node training
            # our cluster sets TASKS_PER_NODE not NTASKS_PER_NODE
            # (with srun, SLURM_STEP_TASKS_PER_NODE)
            # https://slurm.schedmd.com/srun.html#OPT_SLURM_STEP_TASKS_PER_NODE
            if 'SLURM_NTASKS_PER_NODE' not in os.environ.keys():
                os.environ['SLURM_NTASKS_PER_NODE'] = os.environ['SLURM_TASKS_PER_NODE']
                # from lightning_lite.plugins.environments import SLURMEnvironment
                from zoobot.pytorch import manchester
                logging.warning('Using custom slurm environment')
                # https://pytorch-lightning.readthedocs.io/en/stable/clouds/cluster_advanced.html#enable-auto-wall-time-resubmitions
                plugins = [manchester.ManchesterEnvironment(auto_requeue=False)]

    if gpus > 0:
        accelerator = 'gpu'
        devices = gpus
    else:
        accelerator = 'cpu'
        devices = None  # all

    precision = 32
    if mixed_precision:
        logging.info(
            'Training with automatic mixed precision. Will reduce memory footprint but may cause training instability for e.g. resnet')
        precision = 16

    assert num_workers > 0

    if (gpus is not None) and (num_workers * gpus > os.cpu_count()):
        logging.warning(
            """num_workers * gpu > num cpu.
            You may be spawning more dataloader workers than you have cpus, causing bottlenecks.
            Suggest reducing num_workers."""
        )
    if num_workers > os.cpu_count():
        logging.warning(
            """num_workers > num cpu.
            You may be spawning more dataloader workers than you have cpus, causing bottlenecks.
            Suggest reducing num_workers."""
        )

    if catalog is not None:
        assert train_catalog is None
        assert val_catalog is None
        assert test_catalog is None
        catalogs_to_use = {
            'catalog': catalog
        }
    else:
        assert catalog is None
        catalogs_to_use = {
            'train_catalog': train_catalog,
            'val_catalog': val_catalog,
            'test_catalog': test_catalog  # may be None
        }

    if wandb_logger is not None:
        wandb_logger.log_hyperparams({
            'random_state': random_state,
            'epochs': epochs,
            'accelerator': accelerator,
            'gpus': gpus,
            'nodes': nodes,
            'precision': precision,
            'batch_size': batch_size,
            'greyscale': not color,
            'crop_scale_bounds': crop_scale_bounds,
            'crop_ratio_bounds': crop_ratio_bounds,
            'resize_after_crop': resize_after_crop,
            'num_workers': num_workers,
            'prefetch_factor': prefetch_factor,
            'framework': 'pytorch'
        })

    datamodule = GalaxyDataModule(
        label_cols=schema.label_cols,
        # can take either a catalog (and split it), or a pre-split catalog
        **catalogs_to_use,
        # augmentations parameters
        greyscale=not color,
        crop_scale_bounds=crop_scale_bounds,
        crop_ratio_bounds=crop_ratio_bounds,
        resize_after_crop=resize_after_crop,
        # hardware parameters
        batch_size=batch_size, # on 2xA100s, 256 with DDP, 512 with distributed (i.e. split batch)
        num_workers=num_workers,
        prefetch_factor=prefetch_factor
    )
    datamodule.setup(stage='fit')

    # these args are automatically logged
    lightning_model = define_model.ZoobotTree(
        output_dim=len(schema.label_cols),
        question_index_groups=schema.question_index_groups,
        architecture_name=architecture_name,
        channels=channels,
        use_imagenet_weights=False,
        test_time_dropout=True,
        dropout_rate=dropout_rate,
        learning_rate=learning_rate,
        timm_kwargs={'drop_path_rate': drop_connect_rate},
        betas=betas,
        weight_decay=weight_decay,
        scheduler_params=scheduler_params
    )
    
    extra_callbacks = extra_callbacks if extra_callbacks else []

    # used later for checkpoint_callback.best_model_path
    checkpoint_callback = ModelCheckpoint(
            dirpath=os.path.join(save_dir, 'checkpoints'),
            monitor="validation/epoch_loss",
            save_weights_only=True,
            mode='min',
            # custom filename for checkpointing due to / in metric
            filename=checkpoint_file_template,
            # https://pytorch-lightning.readthedocs.io/en/stable/api/pytorch_lightning.callbacks.ModelCheckpoint.html#pytorch_lightning.callbacks.ModelCheckpoint.params.auto_insert_metric_name
            # avoids extra folders from the checkpoint name
            auto_insert_metric_name=auto_insert_metric_name,
            save_top_k=save_top_k
    )

    early_stopping_callback = EarlyStopping(monitor='validation/epoch_loss', patience=patience, check_finite=True)

    callbacks = [checkpoint_callback, early_stopping_callback] + extra_callbacks

    trainer = pl.Trainer(
        log_every_n_steps=150,  # at batch 512 (A100 MP max), DR5 has ~161 train steps
        accelerator=accelerator,
        devices=devices,  # per node
        num_nodes=nodes,
        strategy=strategy,
        precision=precision,
        logger=wandb_logger,
        callbacks=callbacks,
        max_epochs=epochs,
        default_root_dir=save_dir,
        plugins=plugins
    )

    logging.info((trainer.strategy, trainer.world_size,
                 trainer.local_rank, trainer.global_rank, trainer.node_rank))

    trainer.fit(lightning_model, datamodule)  # uses batch size of datamodule

    test_trainer = pl.Trainer(
        accelerator=accelerator,
        devices=1,
        precision=precision,
        logger=wandb_logger,
        default_root_dir=save_dir
    )

    best_model_path = trainer.checkpoint_callback.best_model_path

    # can test as per the below, but note that datamodule must have a test dataset attribute as per pytorch lightning docs.
    # also be careful not to test regularly, as this breaks train/val/test conceptual separation and may cause hparam overfitting
    if test_catalog is not None:
        logging.info(f'Testing on {checkpoint_callback.best_model_path} with single GPU. Be careful not to overfit your choices to the test data...')
        test_trainer.validate(
            model=lightning_model,
            datamodule=datamodule,
            ckpt_path=checkpoint_callback.best_model_path  # can optionally point to a specific checkpoint here e.g. "/share/nas2/walml/repos/gz-decals-classifiers/results/early_stopping_1xgpu_greyscale/checkpoints/epoch=26-step=16847.ckpt"
        )
        test_trainer.test(
            model=lightning_model,
            datamodule=datamodule,
            ckpt_path=checkpoint_callback.best_model_path  # can optionally point to a specific checkpoint here e.g. "/share/nas2/walml/repos/gz-decals-classifiers/results/early_stopping_1xgpu_greyscale/checkpoints/epoch=26-step=16847.ckpt"
        )

    # explicitly update the model weights to the best checkpoint before returning
    # (assumes only one checkpoint callback, very likely in practice)
    # additional kwargs are passed to re-init the lighting_model automatically
    # more broadly, this allows for tracking hparams
    # https://pytorch-lightning.readthedocs.io/en/stable/common/checkpointing_basic.html#initialize-with-other-parameters
    # to make this work, ZoobotLightningModule can only take "normal" parameters (e.g. not custom objects) so has quite a few args
    logging.info('Returning model from checkpoint: {}'.format(best_model_path))
    define_model.ZoobotTree.load_from_checkpoint(best_model_path)  # or .best_model_path, eventually

    return lightning_model, trainer




    # # you can do this to see images, but if you do, wandb will cause training to silently hang before starting if you do this on multi-GPU runs
    # TODO refactor into datamodule setup hook that's only called on main process
    # if wandb_logger is not None:
    #   for (dataloader_name, dataloader) in [('train', datamodule.train_dataloader()), ('val', datamodule.val_dataloader()), ('test', datamodule.test_dataloader())]:
    #     for images, labels in dataloader:
    #       logging.info(images.shape)
    #       images_np = np.transpose(images[:5].numpy(), axes=[0, 2, 3, 1])  # BCHW to BHWC
    #       # images_np = images.numpy()
    #       logging.info((dataloader_name, images_np.shape, images[0].min(), images[0].max()))
    #       wandb_logger.log_image(key="example_{}_images".format(dataloader_name), images=[im for im in images_np[:5]])
    #       break  # only inner loop aka don't log the whole dataloader


def slurm_debugging_logs():
    # https://hpcc.umd.edu/hpcc/help/slurmenv.html
    # logging.info(os.environ)
    logging.info(os.getenv("SLURM_JOB_ID", 'No SLURM_JOB_ID'))
    logging.info(os.getenv("SLURM_JOB_NAME", 'No SLURM_JOB_NAME'))
    logging.info(os.getenv("SLURM_NTASKS", 'No SLURM_NTASKS'))
    # https://github.com/PyTorchLightning/pytorch-lightning/blob/d5fa02e7985c3920e72e268ece1366a1de96281b/pytorch_lightning/trainer/connectors/slurm_connector.py#L29
    # disable slurm detection by pl
    # this is not necessary for single machine, but might be for multi-node
    # may help stop tasks getting left on gpu after slurm exit?
    # del os.environ["SLURM_NTASKS"]  # only exists if --ntasks specified

    logging.info(os.getenv("NODE_RANK", 'No NODE_RANK'))
    logging.info(os.getenv("LOCAL_RANK", 'No LOCAL_RANK'))
    logging.info(os.getenv("WORLD_SIZE", 'No WORLD_SIZE'))



import numpy as np
from scipy.stats import beta


def expected_value_of_dirichlet_mixture(concentrations, schema):
    # badly named vs posteriors, actually gives predicted vote fractions of answers...
    # mean probability (including dropout) of an answer being given. 
    # concentrations has (batch, answer, ...) shape where extra dims are distributions (e.g. galaxies, answer, model, pass)

    # collapse trailing dims
    concentrations = concentrations.reshape(concentrations.shape[0], concentrations.shape[1], -1)

    p_of_answers = []
    for q in schema.questions:
        concentrations_by_q = concentrations[:, q.start_index:q.end_index+1]  # trailing dims are distributions
        for answer_index in range(len(q.answers)):
            # first mean is per distribution
            # second .mean() is average p for all distributions
            mean_of_each_distribution = get_beta_mean(concentrations=concentrations_by_q, answer_index=answer_index)  # (galaxies, distributions)
            # print(mean_of_each_distribution.shape)
            p_of_answers.append(mean_of_each_distribution.mean(axis=1))  # now (galaxies)

    p_of_answers = np.stack(p_of_answers, axis=1)
    return p_of_answers


def get_beta_mean(concentrations, answer_index):
    # concentrations shape (galaxy, answer, distribution)
    concentrations_a = concentrations[:, answer_index]
    concentrations_sum = concentrations.sum(axis=1)
    return concentrations_a / concentrations_sum   # (galaxy, distribution)



# Prob given q is asked
# get prob_of_answers = expected_value_of_dirichlet_mixture(concentrations, schema)
def get_expected_votes_ml(prob_of_answers, question, votes_for_base_question: int, schema, round_votes):
    # (send all prob_of_answers not per-question prob, they are all potentially relevant)
    prev_q = question.asked_after
    if prev_q is None:
        expected_votes = np.ones(len(prob_of_answers)) * votes_for_base_question
    else:
        joint_p_of_asked = schema.joint_p(prob_of_answers, prev_q.text)  # prob of getting the answer needed to ask this question
        expected_votes = joint_p_of_asked * votes_for_base_question
    if round_votes:
        return np.round(expected_votes)
    else:
        return expected_votes


def get_expected_votes_human(label_df, question, votes_for_base_question: int, schema, round_votes):
    # might be better called "get galaxies where at least half of humans actually answered that question, in the labels"
    

    answer_fractions = label_df[[a + '_fraction' for a in schema.label_cols]].values
    # some will be nan as fractions are often nan (as 0 of 0)

    prev_q = question.asked_after
    if prev_q is None:
        expected_votes = np.ones(len(label_df)) * votes_for_base_question
    else:
        # prob of getting the answer needed to ask this question - the product of the (perhaps expected, but here, actual) dependent vote fractions
        joint_p_of_asked = schema.joint_p(answer_fractions, prev_q.text)  
        # for humans, its just votes_for_base_question * the product of all the fractions leading to that q
        expected_votes = joint_p_of_asked * votes_for_base_question
    if round_votes:
        return np.round(expected_votes)
    else:
        return expected_votes




# class DirichletEqualMixture():

#     def __init__(self, concentrations) -> None:
        
#         # concentrations should be shape (galaxy, answer, distributions...)

#         

#         self._concentrations = concentrations

#         # create stats object to do all the work
#         self._dist = dirichlet(self._concentrations)

#     def pdf(self, x):
#         return self._dist.pdf(x)

#     def mean(self):
#         return self._dist.mean()



# https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.dirichlet.html


# for single distribution/model
# def get_confidence_interval_around_mode(concentrations, question_indices, answer_index, interval_width=0.95, gridsize=20):
    
#     concentrations_q = concentrations[:, question_indices[0]:question_indices[1]+1]
#     concentrations_a = concentrations[:, answer_index]
#     concentrations_sum = concentrations_q.sum(axis=1)
#     # dirichlet of this or not this is equivalent to beta distribution with concentrations (this, sum_of_not_this)
#     concentrations_not_a = concentrations_sum - concentrations_a
#     # concentrations_a and concentrations_not_a have shape (batch)

#     # find mode
#     mode = (concentrations_a - 1) / (concentrations_a + concentrations_sum - 2)  # (10,)
#     # can broadcast in galaxy/batch dimension by default!
#     dist = beta(a=concentrations_a, b=concentrations_not_a)
#     cdf_at_mode = dist.cdf(mode)

#     # find what cdf value corresponds to 0.95 width interval around the mode
#     lower_cdf_value = cdf_at_mode - interval_width/2.
#     upper_cdf_value = cdf_at_mode + interval_width/2.

#     # account for edges - replace other bound with interval_width
#     # if upper cdf value would have to be above 1 to include 0.95 mass, set lower bound to e.g. cdf of 0.05
#     upper_bound_above_1 = upper_cdf_value >= 1.
#     lower_cdf_value = np.where(upper_bound_above_1, 1-interval_width, lower_cdf_value)
#     # and set upper bound to exactly 1 / index -1 if needed, below

#     # similarly for lower bound - set upper bound to e.g. cdf of 0.95
#     lower_bound_below_0 = lower_cdf_value <= 0.
#     upper_cdf_value = np.where(lower_bound_below_0, interval_width, upper_cdf_value)

#     # find what x points (vote fractions) have those cdf values
#     # can broadcast in bound dimension with a reshape!
#     dist_with_extra_dim = beta(a=concentrations_a.reshape(-1, 1), b=concentrations_not_a.reshape(-1, 1))
#     grid = np.linspace(0., 1., num=gridsize).reshape(1, -1)  # galaxy, test_cdf_value
#     cdf_grid = dist_with_extra_dim.cdf(grid)
#     # now work out which indices have those values
#     index_of_lower = np.argmin(np.abs(cdf_grid - lower_cdf_value.reshape(-1, 1)), axis=1)
#     index_of_higher = np.argmin(np.abs(cdf_grid - upper_cdf_value.reshape(-1, 1)), axis=1)

#     return grid.squeeze()[index_of_lower], grid.squeeze()[index_of_higher]


def get_confidence_intervals(concentrations, schema, interval_width=.9, gridsize=100):

    # collapse trailing dims
    concentrations = concentrations.reshape(concentrations.shape[0], concentrations.shape[1], -1)

    lower_edges = []
    upper_edges = []
    for q in schema.questions:
        concentrations_q = concentrations[:, q.start_index:q.end_index+1]
        for answer_index in range(len(q.answers)):
            # grid, pdf, cdf = beta_mixture_on_grid(concentrations_q, answer_index, gridsize=gridsize)
            # lower_edge, upper_edge = get_confidence_interval_from_binned_dist(grid, pdf, cdf, interval_width=interval_width)
            lower_edge, upper_edge = get_confidence_interval_from_ppf_medians(concentrations_q, answer_index, interval_width=interval_width)
            lower_edges.append(lower_edge)
            upper_edges.append(upper_edge)

    lower_edges = np.stack(lower_edges, axis=1)
    upper_edges = np.stack(upper_edges, axis=1)

    return lower_edges, upper_edges


def get_confidence_interval_from_ppf_medians(concentrations_q, answer_index, interval_width=.9):
    concentrations_a, concentrations_not_a = reshape_concentrations_for_scipy_beta(concentrations_q, answer_index)
    dist = beta(a=concentrations_a, b=concentrations_not_a)
    lower_edges = dist.ppf(0.5 - interval_width/2.)  # dimension (distribution, galaxy)
    upper_edges = dist.ppf(0.5 + interval_width/2.)
    return np.median(lower_edges, axis=0), np.median(upper_edges, axis=0)  # median over mixture, per galaxy



# supports trailing dimensions for more distributions
def beta_mixture_on_grid(concentrations_q, answer_index, gridsize=100):
    # concentration (galaxy, answer_index, distribution), any extra distribution dims already flattened

    concentrations_a, concentrations_not_a = reshape_concentrations_for_scipy_beta(concentrations_q, answer_index)
    # (distribution, galaxy)

    dist_with_extra_dim = beta(a=np.expand_dims(concentrations_a, -1), b=np.expand_dims(concentrations_not_a, -1))

    grid = np.linspace(1e-8, 1. - 1e-8, num=gridsize).reshape(1, -1)  # galaxy, test_cdf_value
    pdf_grid = dist_with_extra_dim.pdf(grid)
    # normalise over grid dimension
    pdf_grid = pdf_grid / np.sum(pdf_grid, axis=2, keepdims=True)
    # (distribution, galaxy, grid) 

    # take mean over distribution dim (0th)
    mean_pdf = pdf_grid.mean(axis=0)
    # (galaxy, grid)

    # convert to cdf via rolling sum over grid axis
    cdf = np.cumsum(mean_pdf, axis=1)

    return grid.squeeze(), mean_pdf, cdf


def get_confidence_interval_from_binned_dist(grid, pdf, cdf, interval_width=.95):  # grid (100), pdf/cdf (4, 100)

    expected_value = np.sum(grid.reshape(1, -1) * pdf, axis=1)  # (galaxies)
    # print(expected_value.shape)

    loc_of_expected = np.argmin(np.abs(grid.reshape(1, -1) - expected_value.reshape(-1, 1)), axis=1)
    # (galaxy), values of grid_loc (aimed at grid dimension)

    # iterating through the ith element of loc_of_expected, k, we want to select [i, k] from cdf
    # can do this with multi-dim integer array indexing
    # https://numpy.org/devdocs/user/basics.indexing.html#integer-array-indexing
    grid_indices = np.arange(len(loc_of_expected)), loc_of_expected   # NOT a list, or broadcasts to extra dim
    cdf_at_loc_of_expected = cdf[grid_indices]

    lower_cdf_value = cdf_at_loc_of_expected - interval_width/2.
    upper_cdf_value = cdf_at_loc_of_expected + interval_width/2.

    # account for edges - replace other bound with interval_width
    # if upper cdf value would have to be above 1 to include 0.95 mass, set lower bound to e.g. cdf of 0.05
    upper_bound_above_1 = upper_cdf_value >= 1.
    lower_cdf_value = np.where(upper_bound_above_1, 1-interval_width, lower_cdf_value)
    # and set upper bound to exactly 1 / index -1 if needed, below

    # similarly for lower bound - set upper bound to e.g. cdf of 0.95
    lower_bound_below_0 = lower_cdf_value <= 0.
    upper_cdf_value = np.where(lower_bound_below_0, interval_width, upper_cdf_value)

    # now work out which indices have those values
    index_of_lower = np.argmin(np.abs(cdf - lower_cdf_value.reshape(-1, 1)), axis=1)
    index_of_higher = np.argmin(np.abs(cdf - upper_cdf_value.reshape(-1, 1)), axis=1)

    # each of shape (galaxy)
    return grid.squeeze()[index_of_lower], grid.squeeze()[index_of_higher]



def reshape_concentrations_for_scipy_beta(concentrations_q, answer_index):
    # reshape to have distribution in leading dim
    concentrations_q = concentrations_q.transpose(2, 0, 1)
    # (distribution, galaxy, answer_index)

    concentrations_a = concentrations_q[:, :, answer_index]
    concentrations_q_sum = concentrations_q.sum(axis=2)
    # dirichlet of this or not this is equivalent to beta distribution with concentrations (this, sum_of_not_this)
    concentrations_not_a = concentrations_q_sum - concentrations_a

    return concentrations_a, concentrations_not_a



def test_beta_cdf_on_grid():
    question_indices = [0, 2]
    answer_index = 1
    concentrations = np.expand_dims(np.array([[2., 2., 4.], [4., 4., 2.], [4., 4., 2.], [4., 4., 2.]]), axis=2)  # (galaxy=4, answer=3, distribution=1)


    grid, pdf, cdf = beta_mixture_on_grid(concentrations, question_indices, answer_index, gridsize=1000)
    print(grid.shape, pdf.shape, cdf.shape)

    lower_fracs, upper_fracs = get_confidence_interval_from_binned_dist(grid, pdf, cdf)

    with np.printoptions(precision=3):
        print(lower_fracs)
        print(upper_fracs)

#     question_indices = [0, 2]
#     answer_index = 1
#     concentrations = np.array([[2., 2., 4.], [4., 4., 2.]])  # 2 events/galaxies, 3 possible outcomes/answers
#     # should collapse to [[2., 6.], [4., 6.]] in beta [this, not-this] for answer 1

#     # https://homepage.divms.uiowa.edu/~mbognar/applets/beta.html
#     # mode 0.25, 0.4
#     # if lower bound 0: upper bound 0.5207, 0.65506
#     # works well for large gridsize (>100)

# def test_dirichlet_mixture():
#     concentrations = np.expand_dims(np.array([[2., 2., 4.], [4., 4., 2.], [4., 4., 2.], [4., 4., 2.]]), axis=2)

#     mixture = DirichletEqualMixture(concentrations)
#     print(mixture.mean())

def test_get_confidence_interval_from_ppf_medians():
    # question_indices = [0, 2]
    answer_index = 1
    concentrations = np.array([[2., 2., 4.], [4., 4., 2.]])
    # concentrations = np.expand_dims(concentrations, axis=2)
    concentrations = np.stack([concentrations] * 5, axis=2)
    print(concentrations.shape)
    lower_edge, upper_edge = get_confidence_interval_from_ppf_medians(concentrations, answer_index)
    print(lower_edge, upper_edge)


if __name__ == '__main__':

#     # test_beta_cdf_on_grid()

#     test_dirichlet_mixture()

    test_get_confidence_interval_from_ppf_medians()


"""
AlexNet Keras Implementation
BibTeX Citation:
@inproceedings{krizhevsky2012imagenet,
  title={Imagenet classification with deep convolutional neural networks},
  author={Krizhevsky, Alex and Sutskever, Ilya and Hinton, Geoffrey E},
  booktitle={Advances in neural information processing systems},
  pages={1097--1105},
  year={2012}
}
Courtesy https://github.com/eweill/keras-deepcv/blob/master/models/classification/alexnet.py
"""

# Import necessary components to build LeNet
import tensorflow as tf
from tensorflow.keras import layers, regularizers
# from keras.layers.normalization import layers.BatchNormalization


def alexnet_model(img_shape=(224, 224, 1), n_classes=10, l2_reg=0.,
	weights=None):

	# Initialize model
	alexnet = tf.keras.Sequential()

	# Layer 1
	alexnet.add(layers.Conv2D(96, (11, 11), input_shape=img_shape,
		padding='same', kernel_regularizer=regularizers.l2(l2_reg)))
	alexnet.add(layers.BatchNormalization())
	alexnet.add(layers.Activation('relu'))
	alexnet.add(layers.MaxPooling2D(pool_size=(2, 2)))

	# Layer 2
	alexnet.add(layers.Conv2D(256, (5, 5), padding='same'))
	alexnet.add(layers.BatchNormalization())
	alexnet.add(layers.Activation('relu'))
	alexnet.add(layers.MaxPooling2D(pool_size=(2, 2)))

	# Layer 3
	alexnet.add(layers.ZeroPadding2D((1, 1)))
	alexnet.add(layers.Conv2D(512, (3, 3), padding='same'))
	alexnet.add(layers.BatchNormalization())
	alexnet.add(layers.Activation('relu'))
	alexnet.add(layers.MaxPooling2D(pool_size=(2, 2)))

	# Layer 4
	alexnet.add(layers.ZeroPadding2D((1, 1)))
	alexnet.add(layers.Conv2D(1024, (3, 3), padding='same'))
	alexnet.add(layers.BatchNormalization())
	alexnet.add(layers.Activation('relu'))

	# Layer 5
	# alexnet.add(layers.ZeroPadding2D((1, 1)))
	# alexnet.add(layers.Conv2D(1024, (3, 3), padding='same'))
	# alexnet.add(layers.BatchNormalization())
	# alexnet.add(layers.Activation('relu'))
	# alexnet.add(layers.MaxPooling2D(pool_size=(2, 2)))

	# Layer 6
	alexnet.add(layers.Flatten())
	alexnet.add(layers.Dense(3072))
	alexnet.add(layers.BatchNormalization())
	alexnet.add(layers.Activation('relu'))
	alexnet.add(layers.Dropout(0.5))

	# Layer 7
	# alexnet.add(layers.Dense(4096))
	# alexnet.add(layers.BatchNormalization())
	# alexnet.add(layers.Activation('relu'))
	# alexnet.add(layers.Dropout(0.5))

	# Layer 8
	alexnet.add(layers.Dense(n_classes))
	alexnet.add(layers.BatchNormalization())
	alexnet.add(layers.Activation('softmax'))

	if weights is not None:
		alexnet.load_weights(weights)

	return alexnet



import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

# from zoobot.tensorflow.stats import mixture_stats


def dirichlet_prob_of_answers(concentrations, schema):
    # badly named vs posteriors, actually gives predicted vote fractions of answers...
    # mean probability (including dropout) of an answer being given. 
    # concentrations has (batch, answer, ...) shape where extra dims are distributions (e.g. galaxies, answer, model, pass)
    p_of_answers = []
    for q in schema.questions:
        concentrations_by_q = concentrations[:, q.start_index:q.end_index+1, :]
        p_of_answers.append(DirichletMultinomialEqualMixture(total_votes=1, concentrations=concentrations_by_q).mean().numpy())

    p_of_answers = np.concatenate(p_of_answers, axis=1)
    return p_of_answers

class EqualMixture():

    def __init__(self):
        """
        Mixture of abstract distributions.
        Abstract class implementing broadcasting along the final axis (which model) for common funcs e.g. log_prob, mean_log_prob etc

        Raises:
            NotImplementedError: Abstract class!
        """
        raise NotImplementedError

    @property
    def batch_shape(self):
        return self.distributions[0].batch_shape

    @property
    def event_shape(self):
        return self.distributions[0].event_shape
    
    def log_prob(self, x):
        log_probs_by_dist = [dist.log_prob(x) for dist in self.distributions]
        return tf.transpose(log_probs_by_dist)
    
    def prob(self, x):  # just the same, but prob not log prob
        return tf.math.exp(self.log_prob(x))

    def mean_prob(self, x):
        # expected value across all distributions - very handy
        prob = self.prob(x)
        return tf.reduce_mean(prob, axis=-1)

    def mean_log_prob(self, x):
        return tf.math.log(self.mean_prob(x))

    def mean(self):
        # the mean of the mean (expected) value of each distribution
        # aka the mean of the expected vote fractions
        return tf.reduce_mean([dist.mean() for dist in self.distributions], axis=0)

    def mean_cdf(self, x):  # as integration is separable 
        # will fail as cdf is not implemented
        return tf.reduce_mean([dist.cdf(x) for dist in self.distributions], axis=0)

    def cdf(self, x):
        cdf_by_dist = [dist.cdf(x) for dist in self.distributions]
        return tf.transpose(cdf_by_dist)
        # raise NotImplementedError('Can only be calculated with batch_dim shaped x, unlike .prob etc, needs custom implementation')



class DirichletEqualMixture(EqualMixture):

    def __init__(self, concentrations):
        """
        Equally-weighted mixture of Dirichlet distributions.
        Basically, the Dirichlet version of a Gaussian mixture model.
        Useful for combining predictions from multiple models or multiple forward passes (with MC Dropout and augmentations).

        Specifically, self.distributions will be a list of tfp.distributions.Dirichlet, each with concentrations (galaxies, answers, specific model_index)
        Broadcasting works along galaxies dimension by subclassing ``EqualMixture``.

        Args:
            concentrations (np.ndarray): predicted dirichlet parameters for one galaxy, of shape (galaxies, answers, models)
        """
        # flatten trailing dims (e.g. galaxies, answer, model, pass) to (galaxies, answer, distribution)
        concentrations = concentrations.reshape(concentrations.shape[0], concentrations.shape[1], -1)
        concentrations = concentrations.astype(np.float32)

        self.concentrations = concentrations

        self.n_distributions = self.concentrations.shape[2]
        self.distributions = [
            tfp.distributions.Dirichlet(self.concentrations[:, :, n], validate_args=True)
            for n in range(self.n_distributions)
        ]

    # def entropy_estimate(self):
    #     """
    #     Returns:
    #         np.ndarray: midpoint between entropy bounds
    #     """
    #     upper = self.entropy_upper_bound()
    #     lower = self.entropy_lower_bound()
    #     return lower + (upper-lower)/2.  # midpoint between bounds

    # def entropy_upper_bound(self):
    #     """
    #     Returns:
    #         np.ndarray: upper bound on entropy. See ``mixture_stats.entropy_upper_bound``.
    #     """
    #     return np.array([mixture_stats.entropy_upper_bound(galaxy_conc, weights=np.ones(self.n_distributions)/self.n_distributions) for galaxy_conc in self.concentrations])

    # def entropy_lower_bound(self):
    #     """
    #     Returns:
    #         np.ndarray: lower bound on entropy. See ``mixture_stats.entropy_lower_bound``.
    #     """
    #     return np.array([mixture_stats.entropy_lower_bound(galaxy_conc, weights=np.ones(self.n_distributions)/self.n_distributions) for galaxy_conc in self.concentrations])


class DirichletMultinomialEqualMixture(EqualMixture):

    def __init__(self, total_votes, concentrations, *args, **kwargs):
        """

        Mixture of equally-weighted Dirichlet-Multinomial distributions.
        Useful for interpreting the aggregate predictions of many models in terms of expected volunteer responses.

        Uses the same interface as tfp, but can aggregate across models (by subclassing ``EqualMixture``).

        Args:
            total_votes (np.ndarray): total votes for some question, of shape (galaxies)
            concentrations (np.ndarray): of shape (galaxies, answer, ...). Answers for *one question only* due to total votes. Extra dims flattened.
        """
        
        self.total_votes = np.array(total_votes).astype(np.float32)

        # flatten trailing dims (e.g. galaxies, answer, model, pass) to (galaxies, answer, distribution)
        concentrations = concentrations.reshape(concentrations.shape[0], concentrations.shape[1], -1)
        self.concentrations = concentrations.astype(np.float32)

        self.n_distributions = self.concentrations.shape[2]
        self.distributions = [
            tfp.distributions.DirichletMultinomial(self.total_votes, self.concentrations[:, :, n], validate_args=True)
            for n in range(self.n_distributions)
        ]



    def to_beta(self, answer_index, batch_dim):
        """
        Convert from Dirichlet (multivariate) to Beta (bivariate) by adding up all the parameters to every answer but one.
        E.g. 3-variate dirichlet(concentrations=[2, 3, 6]) converts to bivariate beta(a=2, b=3+6) for ``answer_index=0``
        Intuitively, this is like going from "is it a, b, or c?" to "is it a or not-a?"

        Args:
            answer_index (int): which answer to use as first concentration. Other answers will have concentrations summed.
            batch_dim (int): batch dimension i.e. num dirichlet distributions in this mixture.

        Returns:
            BetaEqualMixture: beta(answer index, not answer index) with a batch dimension of num. dirichlet distributions in this mixture
        """
        assert self.batch_shape == 1  # beta uses batch shape for cdf(x), so needs to be able to broadcast, so can only do one galaxy at a time
        num_answers = self.concentrations.shape[1]
        answer_concentrations = self.concentrations[:, answer_index]
        all_concentrations = self.concentrations.reshape(1, num_answers, -1).sum(axis=1)
        beta_concentrations = np.stack([answer_concentrations, all_concentrations - answer_concentrations], axis=1)
        return BetaEqualMixture(beta_concentrations, batch_dim=batch_dim)



class BetaEqualMixture(DirichletEqualMixture):

    def __init__(self, concentrations):
        """
        Batch dim is the number of models in the mixture, 

        Only supports galaxy_dim = 1 i.e. one galaxy at a time.
    
        As with DirichletEqualMixture, self.distributions is list of tfp.distributions.Beta, each with dimension (galaxy, answer)

        Only used for calculating confidence intervals for a specific answer - which is currently impractically slow anyway.

        Args:
            concentrations ([type]): of shape (1, 2, ...) for 1 galaxy, 2 answers, and trailing distibution dims
        """
        assert concentrations.shape[0] == 1  # 1 galaxy
        assert concentrations.shape[1] == 2  # 2 answers (i.e. "this answer" or "not this answer")
        # remaining dimensions assumed to be distributions
        concentrations = concentrations.reshape(concentrations.shape[0], concentrations.shape[1], -1)
        # print(concentrations.shape)
        
        self.batch_dim = batch_dim
        self.concentrations = np.stack([np.squeeze(concentrations) for _ in range(batch_dim)], axis=0).astype(np.float32)
        # print(self.concentrations.shape)
        assert self.concentrations.ndim == 3
        self.n_distributions = self.concentrations.shape[2]
        self.distributions = [
            tfp.distributions.Beta(
                concentration0=self.concentrations[:, 0, n],
                concentration1=self.concentrations[:, 1, n],
                validate_args=True)
            for n in range(self.n_distributions)
        ]

    @property
    def standard_sample_grid(self):
        return np.linspace(1e-8, 1. - 1e-8, num=self.batch_dim)

    def mean_mode(self):  # approximate only
        assert self.batch_dim >= 50
        # print(self.mean_prob(self.standard_sample_grid).shape)
        mode_index = np.argmax(self.mean_prob(self.standard_sample_grid))
        return self.standard_sample_grid[mode_index]  # bit lazy but it works

    def confidence_interval(self, interval_width):
        assert self.batch_dim >= 50
        # dist must be unimodal (though the mode can be at extremes)
        cdf = self.mean_cdf(self.standard_sample_grid)
        mode = self.mean_mode()
        # print(mode)
        mode_cdf = self.mean_cdf(mode)[0]  # just using the first, batch size broadcasting gives batch_size identical results
        # print(mode_cdf)
        return confidence_interval_from_cdf(self.standard_sample_grid, cdf, mode_cdf, interval_width)

    def cdf(self, x):
        # for dist in self.distributions:
            # dist_with_batch_matching_x = tfp.distributions.Beta()
        # print(x)
        cdfs = [dist.cdf(x) for dist in self.distributions]
        return tf.transpose(cdfs)


def confidence_interval_from_cdf(x, cdf, mode_cdf, interval_width):
    width_per_half = interval_width / 2

    lower_index = np.argmin(np.abs(mode_cdf  - cdf - width_per_half))
    upper_index = np.argmin(np.abs(cdf - mode_cdf - width_per_half))

    if mode_cdf < width_per_half:  # i.e. if lower interval hits the edge
        upper_index = np.argmin(np.abs(cdf - interval_width))
    elif 1 - mode_cdf < width_per_half:  # i.e. upper interval hits the edge
        lower_index = np.argmin(np.abs(1 - cdf - interval_width))

    assert np.allclose(cdf[upper_index] - cdf[lower_index], interval_width, atol=.02)

    lower = x[lower_index]
    upper = x[upper_index]
    return lower, upper




# if __name__ == '__main__':


    # total_votes = 10
    # votes_except_last = np.random.randint(0, int(total_votes/2), size=(10, 2))  # can be no more than half, for convenience
    # votes_last = total_votes - votes_except_last.sum(axis=1)
    # votes = np.concatenate([votes_except_last, votes_last.reshape(-1, 1)], axis=1)

    # concentrations = np.random.rand(10, 3, 2, 5) * 100 + 1  # 10 events/galaxies, 3 possible outcomes/answers, 2 models, 5 forward passes

    # mixture = DirichletMultinomialEqualMixture(total_votes=total_votes, concentrations=concentrations)
    # print(mixture.mean())
    # print(mixture.mean().shape)  # taking a mean across distributions. not totally sure this is correct approach

    # eval_points_except_last = np.random.rand(*concentrations.shape) / 3
    # eval_points_last = 1 - eval_points_except_last.sum(axis=1)
    # eval_points = np.concatenate([eval_points_except_last, eval_points_last.reshape(-1, 1)], axis=1)

    # mixture = DirichletEqualMixture(concentrations=concentrations)

    # print(mixture.cdf(eval_points))

    # mixture = tfp.distributions.Beta(5., 6.)  # concentrations
    # print(mixture.cdf(.5))

    # for each galaxy, for each question, for each answer, create a Beta distribution for that answer or not-that-answer
    # then get the cdf
    # no need for TFP - probably WAY faster in pure python
    # I just want the CDF



# # only used for posthoc evaluation, not when training
# def dirichlet_mixture_loss(labels, predictions, question_index_groups):  # pasted
#     q_losses = []
#     for q_n in range(len(question_index_groups)):
#         q_indices = question_index_groups[q_n]
#         q_start = q_indices[0]
#         q_end = q_indices[1]
#         q_loss = dirichlet_mixture_loss_per_question(labels[:, q_start:q_end+1], predictions[:, q_start:q_end+1])
#         q_losses.append(q_loss)
    
#     total_loss = np.stack(q_losses, axis=1)
#     return total_loss  # leave the reduce_sum to the estimator

# # this works but is very slow
# # def dirichlet_mixture_loss_per_question(labels_q, predictions_q):
# #     n_samples = predictions_q.shape[-1]
# #     total_votes = labels_q.sum(axis=1).squeeze()
# #     log_probs = []
# # #     print(predictions_q.shape,total_votes.shape, n_samples)
# #     for n, galaxy in enumerate(predictions_q):
# #         mixture = acquisition_utils.dirichlet_mixture(np.expand_dims(galaxy, axis=0), total_votes[n], n_samples)
# #         log_probs.append(mixture.log_prob(labels_q[n]))
# #     return -np.squeeze(np.array(log_probs))  # negative log prob

# def dirichlet_mixture_loss_per_question(labels_q, concentrations_q):
#     n_samples = concentrations_q.shape[-1]
#     total_votes = labels_q.sum(axis=1).squeeze()
#     mean_log_probs = DirichletMultinomialEqualMixture(total_votes=total_votes, concentrations=concentrations_q).mean_log_prob(labels_q) 
#     return -np.squeeze(np.array(mean_log_probs))  # negative log prob


# def beta_confidence_interval(concentration0, concentration1, interval_width):

#     dist = tfp.distributions.Beta(concentration0, concentration1, validate_args=True, allow_nan_stats=False)
#     if concentration0 <= 1:
#         mode = 1
#     elif concentration1 <= 1:
#         mode = 0
#     else:
#         mode = dist.mode()
#     mode_cdf = dist.cdf(mode)

#     x = np.linspace(0.001, .999, num=1000)
#     cdf = dist.cdf(x)
#     return confidence_interval_from_cdf(x, cdf, mode_cdf, interval_width)



# Based on Inigo's BYOL FT step
# https://github.com/inigoval/finetune/blob/main/finetune.py
import logging
import os
import warnings
from functools import partial

import pytorch_lightning as pl
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint

import torch
import torch.nn.functional as F
import torchmetrics as tm

from zoobot.pytorch.training import losses
from zoobot.pytorch.estimators import define_model
from zoobot.shared import schemas

# https://discuss.pytorch.org/t/how-to-freeze-bn-layers-while-training-the-rest-of-network-mean-and-var-wont-freeze/89736/7
# I do this recursively and only for BatchNorm2d (not dropout, which I still want active)


def freeze_batchnorm_layers(model):
    for name, child in (model.named_children()):
        if isinstance(child, torch.nn.BatchNorm2d):
            logging.debug('freezing {} {}'.format(child, name))
            child.eval()  # no grads, no param updates, no statistic updates
        else:
            freeze_batchnorm_layers(child)  # recurse


class FinetuneableZoobotAbstract(pl.LightningModule):
    """
    Parent class of :class:`FinetuneableZoobotClassifier` and :class:`FinetuneableZoobotTree`.
    You cannot use this class directly - you must use the child classes above instead.

    This class defines the finetuning methods that those child classes both use.
    For example: when provided `checkpoint_loc`, it will load the encoder from that checkpoint.
    Both :class:`FinetuneableZoobotClassifier` and :class:`FinetuneableZoobotTree`
    can (and should) be passed any of these arguments to customise finetuning.

    You could subclass this class to solve new finetuning tasks (like regression) - see :ref:`advanced_finetuning`.

    Args:
        checkpoint_loc (str, optional): Path to encoder checkpoint to load (likely a saved ZoobotTree). Defaults to None.
        encoder (pl.LightningModule, optional): Alternatively, pass an encoder directly. Load with :func:`zoobot.pytorch.training.finetune.load_pretrained_encoder`.
        encoder_dim (int, optional): Output dimension of encoder. Defaults to 1280 (EfficientNetB0's encoder dim).
        lr_decay (float, optional): For each layer i below the head, reduce the learning rate by lr_decay ^ i. Defaults to 0.75.
        weight_decay (float, optional): AdamW weight decay arg (i.e. L2 penalty). Defaults to 0.05.
        learning_rate (float, optional): AdamW learning rate arg. Defaults to 1e-4.
        dropout_prob (float, optional): P of dropout before final output layer. Defaults to 0.5.
        freeze_batchnorm (bool, optional): If True, do not update batchnorm stats during finetuning. Defaults to True.
        prog_bar (bool, optional): Print progress bar during finetuning. Defaults to True.
        visualize_images (bool, optional): Upload example images to WandB. Good for debugging but slow. Defaults to False.
        seed (int, optional): random seed to use. Defaults to 42.
    """

    def __init__(
        self,
        # can provide either checkpoint_loc, and will load this model as encoder...
        checkpoint_loc=None,
        # ...or directly pass model to use as encoder
        encoder=None,
        encoder_dim=1280,  # as per current Zooot. TODO Could get automatically?
        n_epochs=100,  # TODO early stopping
        n_layers=0,  # how many layers deep to FT
        lr_decay=0.75,
        weight_decay=0.05,
        learning_rate=1e-4,
        dropout_prob=0.5,
        freeze_batchnorm=True,
        prog_bar=True,
        visualize_images=False,  # upload examples to wandb, good for debugging
        seed=42
    ):
        super().__init__()

        # adds every __init__ arg to model.hparams
        # will also add to wandb if using logging=wandb, I think
        # necessary if you want to reload!
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # this raises a warning that encoder is already a Module hence saved in checkpoint hence no need to save as hparam
            # true - except we need it to instantiate this class, so it's really handy to have saved as well
            # therefore ignore the warning
            self.save_hyperparameters()

        if checkpoint_loc is not None:
          assert encoder is None, 'Cannot pass both checkpoint to load and encoder to use'
          self.encoder = load_pretrained_encoder(checkpoint_loc)
        else:
          assert checkpoint_loc is None, 'Cannot pass both checkpoint to load and encoder to use'
          self.encoder = encoder

        self.encoder_dim = encoder_dim
        self.n_layers = n_layers
        self.freeze = True if n_layers == 0 else False

        self.learning_rate = learning_rate
        self.lr_decay = lr_decay
        self.weight_decay = weight_decay
        self.dropout_prob = dropout_prob
        self.n_epochs = n_epochs

        self.freeze_batchnorm = freeze_batchnorm

        if self.freeze_batchnorm:
            freeze_batchnorm_layers(self.encoder)  # inplace

        self.seed = seed
        self.prog_bar = prog_bar
        self.visualize_images = visualize_images

    def configure_optimizers(self):

        if self.freeze:
            params = self.head.parameters()
            return torch.optim.AdamW(params, lr=self.learning_rate)
        else:
            lr = self.learning_rate
            params = [{"params": self.head.parameters(), "lr": lr}]

            # this bit is specific to Zoobot EffNet
            encoder_blocks = list(self.encoder.children())

            # for n, l in enumerate(encoder_blocks):
            #     print('\n')
            #     print(n)
            #     print(l)
            
            # layers with no parameters don't count
            tuneable_blocks = [b for b in encoder_blocks if is_tuneable(b)]
 
            assert self.n_layers <= len(
                tuneable_blocks
            ), f"Network only has {len(tuneable_blocks)} tuneable blocks, {self.n_layers} specified for finetuning"

            # Append parameters of layers for finetuning along with decayed learning rate
            blocks_to_tune = tuneable_blocks[:self.n_layers]
            blocks_to_tune.reverse()  # highest block to lowest block
            for i, layer in enumerate(blocks_to_tune):
                params.append({
                    "params": layer.parameters(),
                    "lr": lr * (self.lr_decay**i)
                })

            # Initialize AdamW optimizer
            opt = torch.optim.AdamW(
                params, weight_decay=self.weight_decay, betas=(0.9, 0.999))  # higher weight decay is typically good

            return opt


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x = self.head(x)
        return x

    
    def make_step(self, batch):
      # part of training/val/test for all subclasses
        x, y = batch
        y_pred = self.forward(x)
        loss = self.loss(y_pred, y)
        # {'loss': loss.mean(), 'predictions': y_pred, 'labels': y}
        return y, y_pred, loss


    # def on_train_batch_end(self, outputs, *args) -> None:
    #     self.log("finetuning/train_loss_batch",
    #              outputs['loss'], on_step=False, on_epoch=True, prog_bar=self.prog_bar)

    def train_epoch_end(self, outputs, *args) -> None:
        losses = torch.cat([batch_output['loss'].expand(0)
                           for batch_output in outputs])
        self.log("finetuning/train_loss",
                 losses.mean(), prog_bar=self.prog_bar)

        if hasattr(self, 'train_acc') is not None:
            self.log("finetuning/train_acc", self.train_acc,
                     prog_bar=self.prog_bar)  # log here

    def validation_epoch_end(self, outputs, *args) -> None:
        # calc. mean of losses over val batches as val loss
        losses = torch.FloatTensor([batch_output['loss']
                                   for batch_output in outputs])
        self.log("finetuning/val_loss", torch.mean(losses),
                 prog_bar=self.prog_bar)

        if hasattr(self, 'val_acc'):
            self.log("finetuning/val_acc", self.val_acc,
                     prog_bar=self.prog_bar)

    def on_test_batch_end(self, outputs, *args) -> None:
        self.log('test/test_loss', outputs['loss'])
        if hasattr(self, 'test_acc'):
            self.test_acc(outputs['predictions'], outputs['labels'])
            self.log(f"finetuning/test_acc", self.test_acc,
                     on_step=False, on_epoch=True)

    def on_validation_batch_end(self, outputs, batch, batch_idx, *args) -> None:
        # self.log(f"finetuning/val_loss_batch",
        #          outputs['loss'].mean(), on_step=False, on_epoch=True, prog_bar=self.prog_bar)
        
        if self.visualize_images:
          self.upload_images_to_wandb(outputs, batch, batch_idx)

    def upload_images_to_wandb(self, outputs, batch, batch_idx):
      raise NotImplementedError('Must be subclassed')



class FinetuneableZoobotClassifier(FinetuneableZoobotAbstract):
    """
    Pretrained Zoobot model intended for finetuning on a classification problem.

    You must also pass either ``checkpoint_loc`` (to a saved encoder checkpoint)
    or `encoder` (to a pytorch model already loaded in memory).
    See :class:FinetuneableZoobotAbstract for more options.

    Any args not in the list below are passed to :class:``FinetuneableZoobotAbstract`` (usually to specify how to carry out the finetuning)

    Args:
        num_classes (int): num. of target classes (e.g. 2 for binary classification).
        label_smoothing (float, optional): See torch cross_entropy_loss docs. Defaults to 0.
        
    """

    def __init__(
            self,
            num_classes: int,
            label_smoothing=0.,
            **super_kwargs) -> None:

        super().__init__(**super_kwargs)

        logging.info('Using classification head and cross-entropy loss')
        self.head = LinearClassifier(
            input_dim=self.encoder_dim,
            output_dim=num_classes,
            dropout_prob=self.dropout_prob
        )
        self.label_smoothing = label_smoothing
        self.loss = partial(cross_entropy_loss,
                            label_smoothing=self.label_smoothing)
        self.train_acc = tm.Accuracy(task='binary', average="micro")
        self.val_acc = tm.Accuracy(task='binary', average="micro")
        self.test_acc = tm.Accuracy(task='binary', average="micro")


    def training_step(self, batch, batch_idx):
        y, y_pred, loss = self.make_step(batch)

        # calculate metrics
        y_class_preds = torch.argmax(y_pred, axis=1)
        self.train_acc(y_class_preds, y)  # update calc, but do not log

        return {'loss': loss.mean(), 'predictions': y_pred, 'labels': y}

    
    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        y, y_pred, loss = self.make_step(batch)

        y_class_preds = torch.argmax(y_pred, axis=1)
        self.val_acc(y_class_preds, y)

        return {'loss': loss.mean(), 'predictions': y_pred, 'labels': y}

    
    def predict_step(self, x, batch_idx):
        # assert False
        x = self.forward(x)  # logits from LinearClassifier
        # then applies softmax
        return F.softmax(x, dim=1)[:, 1]


    def upload_images_to_wandb(self, outputs, batch, batch_idx):
      # self.logger is set by pl.Trainer(logger=) argument
        if (self.logger is not None) and (batch_idx == 0):
            x, y = batch
            y_pred_softmax = F.softmax(outputs['predictions'], dim=1)[:, 1]  # odds of class 1 (assumed binary)
            n_images = 5
            images = [img for img in x[:n_images]]
            captions = [f'Ground Truth: {y_i} \nPrediction: {y_p_i}' for y_i, y_p_i in zip(
                y[:n_images], y_pred_softmax[:n_images])]
            self.logger.log_image(
                key='val_images',
                images=images,
                caption=captions)


class FinetuneableZoobotTree(FinetuneableZoobotAbstract):
    """
    Pretrained Zoobot model intended for finetuning on a decision tree (i.e. GZ-like) problem.

    You must also pass either ``checkpoint_loc`` (to a saved encoder checkpoint)
    or ``encoder`` (to a pytorch model already loaded in memory).
    See :class:FinetuneableZoobotAbstract for more options.

    Args:
        schema (schemas.Schema): description of the layout of the decision tree. See :class:`zoobot.shared.schemas.Schema`.
    """

    def __init__(
        self,
        schema: schemas.Schema,
        **super_kwargs
    ):

        super().__init__(**super_kwargs)

        logging.info('Using dropout+dirichlet head and dirichlet (count) loss')

        self.schema = schema
        self.output_dim = len(self.schema.label_cols)

        self.head = define_model.get_pytorch_dirichlet_head(
            encoder_dim=self.encoder_dim,
            output_dim=self.output_dim,
            test_time_dropout=False,
            dropout_rate=self.dropout_prob
        )
      
        self.loss = define_model.get_dirichlet_loss_func(self.schema.question_index_groups)

    def training_step(self, batch, batch_idx):
        y, y_pred, loss = self.make_step(batch)
        return {'loss': loss.mean(), 'predictions': y_pred, 'labels': y}

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
      # now identical to above
        y, y_pred, loss = self.make_step(batch)
        return {'loss': loss.mean(), 'predictions': y_pred, 'labels': y}

    def test_step(self, batch, batch_idx, dataloader_idx=0):
      # now identical to above
        y, y_pred, loss = self.make_step(batch)
        return {'loss': loss, 'predictions': y_pred, 'labels': y}

    def upload_images_to_wandb(self, outputs, batch, batch_idx):
      pass  # not yet implemented

# https://github.com/inigoval/byol/blob/1da1bba7dc5cabe2b47956f9d7c6277decd16cc7/byol_main/networks/models.py#L29
class LinearClassifier(torch.nn.Module):
    def __init__(self, input_dim, output_dim, dropout_prob=0.5):
        # input dim is representation dim, output_dim is num classes
        super(LinearClassifier, self).__init__()
        self.dropout = torch.nn.Dropout(p=dropout_prob)
        self.linear = torch.nn.Linear(input_dim, output_dim)

    def forward(self, x):
        # returns logits, as recommended for CrossEntropy loss
        x = self.dropout(x)
        x = self.linear(x)
        return x


def cross_entropy_loss(y_pred, y, label_smoothing=0.):
    # y should be shape (batch) and ints
    # y_pred should be shape (batch, classes)
    # returns loss of shape (batch)
    # will reduce myself
    return F.cross_entropy(y_pred, y.long(), label_smoothing=label_smoothing, reduction='none')


def dirichlet_loss(y_pred, y, question_index_groups):
    # aggregation equiv. to sum(axis=1).mean(), but fewer operations
    # returns loss of shape (batch)
    # my func uses sklearn convention y, y_pred
    return losses.calculate_multiquestion_loss(y, y_pred, question_index_groups).mean()*len(question_index_groups)


class FinetunedZoobotClassifierBaseline(FinetuneableZoobotClassifier):
    # exactly as the Finetuned model above, but with a simple single learning rate
    # useful for training from-scratch model exactly as if it were finetuned, as a baseline

    def configure_optimizers(self):
        head_params = list(self.head.parameters())
        encoder_params = list(self.encoder.parameters())
        return torch.optim.AdamW(head_params + encoder_params, lr=self.learning_rate)


def load_pretrained_encoder(checkpoint_loc: str) -> torch.nn.Sequential:
    """
    Args:
        checkpoint_loc (str): path to saved LightningModule checkpoint, likely of :class:`ZoobotTree`, :class:`FinetuneableZoobotClassifier`, or :class:`FinetunabelZoobotTree`. Must have .encoder attribute.

    Returns:
        torch.nn.Sequential: pretrained PyTorch encoder within that LightningModule.
    """
    return define_model.ZoobotTree.load_from_checkpoint(
        checkpoint_loc).encoder


def get_trainer(
    save_dir: str,
    file_template="{epoch}",
    save_top_k=1,
    max_epochs=100,
    patience=10,
    devices=None,
    accelerator='auto',
    logger=None,
    **trainer_kwargs
) -> pl.Trainer:
    """
    PyTorch Lightning Trainer that carries out the finetuning process.
    Use like so: trainer.fit(model, datamodule)

    Args:
        save_dir (str): folder in which to save checkpoints and logs.
        file_template (str, optional): custom naming for checkpoint files. See Lightning docs. Defaults to "{epoch}".
        save_top_k (int, optional): save the top k checkpoints only. Defaults to 1.
        max_epochs (int, optional): train for up to this many epochs. Defaults to 100.
        patience (int, optional): wait up to this many epochs for decreasing loss before ending training. Defaults to 10.
        devices (str, optional): number of devices for training (typically, num. GPUs). Defaults to None.
        accelerator (str, optional): which device to use (typically 'gpu' or 'cpu'). Defaults to 'auto'.
        logger (pl.loggers.wandb_logger, optional): If pl.loggers.wandb_logger, track experiment on Weights and Biases. Defaults to None.

    Returns:
        pl.Trainer: PyTorch Lightning trainer object for finetuning a model on a GalaxyDataModule.
    """

    checkpoint_callback = ModelCheckpoint(
        monitor='finetuning/val_loss',
        every_n_epochs=1,
        save_on_train_epoch_end=True,
        auto_insert_metric_name=False,
        verbose=True,
        dirpath=os.path.join(save_dir, 'checkpoints'),
        filename=file_template,
        save_weights_only=True,
        save_top_k=save_top_k
    )

    early_stopping_callback = EarlyStopping(
        monitor='finetuning/val_loss',
        mode='min',
        patience=patience
    )

    # Initialise pytorch lightning trainer
    trainer = pl.Trainer(
        logger=logger,
        callbacks=[checkpoint_callback, early_stopping_callback],
        max_epochs=max_epochs,
        accelerator=accelerator,
        devices=devices,
        **trainer_kwargs,
    )

    return trainer


def is_tuneable(block_of_layers):
    if len(list(block_of_layers.parameters())) == 0:
        logging.info('Skipping block with no params')
        logging.info(block_of_layers)
        return False
    else:
        # currently, allowed to include batchnorm
        return True

    # when ready (don't peek often, you'll overfit)
    # trainer.test(model, dataloaders=datamodule)

    # return model, checkpoint_callback.best_model_path
    # trainer.callbacks[checkpoint_callback].best_model_path?

# def investigate_structure():

#     from zoobot.pytorch.estimators import define_model


#     model = define_model.get_plain_pytorch_zoobot_model(output_dim=1280, include_top=False)

#     # print(model)
#     # with include_top=False, first and only child is EffNet
#     effnet_with_pool = list(model.children())[0]

#     # 0th is actually EffNet, 1st and 2nd are AvgPool and Identity
#     effnet = list(effnet_with_pool.children())[0]

#     for layer_n, layer in enumerate(effnet.children()):
#         # first bunch are Sequential module wrapping e.g. 3 MBConv blocks
#         print('\n', layer_n)
#         if isinstance(layer, torch.nn.Sequential):
#             print(layer)
#     # so the blocks to finetune are each Sequential (repeated MBConv) block
#     # and other blocks can be left alone
#     # (also be careful to leave batch-norm alone)



import math  # not torch, just python math!

import torch
from torch import nn, Tensor


def custom_top_dirichlet(input_dim, output_dim):
    """
    Final dense layer used in GZ DECaLS (after global pooling). 
    ``output_dim`` neurons with an activation of ``tf.nn.sigmoid(x) * 100. + 1.``, chosen to ensure 1-100 output range
    This range is suitable for parameters of Dirichlet distribution.

    Args:
        output_dim (int): Dimension of dense layer e.g. 34 for decision tree with 34 answers

    Returns:
        nn.Sequential: nn.Linear followed by 1-101 sigmoid activation
    """
    return nn.Sequential(
        # LinearWithCustomInit(in_features=input_dim, out_features=output_dim),
        nn.Linear(in_features=input_dim, out_features=output_dim),
        ScaledSigmoid()
    )


class LinearWithCustomInit(nn.Linear):

    # fan_in = in_features e.g. 1280


    # identical __init__ etc

    # https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/linear.py#L103
    def reset_parameters(self) -> None:
        # # https://pytorch.org/docs/stable/nn.init.html#torch.nn.init.kaiming_uniform_
        # # nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        # nn.init.uniform_(self.weight, 2., 3.)
        # if self.bias is not None:
        #     # print(self.weight.size(1))
        #     # fan_in = weight.size(1) = input_dim (e.g. 1280)
        #     fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
        #     bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
        #     # print(self.bias, bound)
        #     # set bias to have values uniformly drawn between -bound and bound (i.e. 1/sqrt(1280))
        #     nn.init.uniform_(self.bias, -bound, bound)

        # glorot init
        fan_in = self.weight.size(1)
        fan_out = self.weight.size(0)
        weight_bound = math.sqrt(6/(fan_in + fan_out))
        # but with extra factor 10 lower
        nn.init.uniform_(self.weight, -weight_bound/10, weight_bound/10)
        # nn.init.zeros_(self.bias)
        # nn.init.uniform_(self.weight, -0.0101, -0.01)
        # nn.init.zeros_(self.bias)
        nn.init.uniform_(self.bias, -4.5, -3.5)

        # super().reset_parameters()  # sets self.bias, self.weight

class ScaledSigmoid(nn.modules.Sigmoid):
    # https://pytorch.org/docs/stable/_modules/torch/nn/modules/activation.html#ReLU

    def forward(self, input: Tensor) -> Tensor:
        """
        Args:
            input (Tensor): any vector. typically logits from a neural network

        Returns:
            Tensor: input mapped to range (1, 101) via torch.sigmoid
        """
        return torch.sigmoid(input) * 100. + 1.  # could make args if I needed


if __name__ == '__main__':
    # print(nn.Linear(1280, 40).weight)
    # print(LinearWithCustomInit(1280, 40).weight)


    x = torch.rand(1280)

    # print(nn.Linear(1280, 40)(x))
    # print(LinearWithCustomInit(1280, 40)(x))

    # print(torch.sigmoid(nn.Linear(1280, 40)(x)))
    # print(torch.sigmoid(LinearWithCustomInit(1280, 40)(x)))


    lin = nn.Linear(1280, 40)(x)
    lin_cust = LinearWithCustomInit(1280, 40)(x)

    print(ScaledSigmoid()(lin))
    print(ScaledSigmoid()(lin_cust))


import logging
from functools import partial
from typing import List

import torch
from torch import nn
import pytorch_lightning as pl
from torchmetrics import Accuracy
import timm

from zoobot.pytorch.estimators import efficientnet_custom, custom_layers
from zoobot.pytorch.training import losses

# overall strategy
# timm for defining complicated pytorch modules
# manual modules for simple things (e.g. heads)
# together, this give pure pytorch self.encoder and self.head
# lightning modules with training instructions built out of those blocks (Zoobot, FinetuneableZooot)
# FinetuneableZoobot could subclass Zoobot to have different optimisers/init args - first has arch stuff, second has optim stuff

# to create from-scratch zoobot model:
# ZoobotTree(pytorch_encoder_args, head_args)   (needs hparam args to be restoreable)
# and within that init:
# pytorch_encoder = get_pytorch_encoder(timm model and model args)
# pytorch_dirichlet_head = get_pytorch_dirichlet_head...
# and then train from scratch

# to finetune:
# load ZoobotTree from checkpoint and keep only ZoobotEncoder
# encoder = ZoobotTree.load_from_checkpoint().encoder   (could wrap in load_pretrained_encoder)
# FinetuneableZoobotClassifier(pretrained_model.encoder, optim_args, task_args)
# (same approach for FinetuneableZoobotTree)

# to use just the encoder later: 
# encoder = load_pretrained_encoder(pyramid=False)
# when pyramid=True, reset the timm model to pull lightning features (TODO)

# timm gives regular pytorch models (with .forward_features argument available)
# for both training and finetuning, we also use some custom torch classes as heads
# DirichletHead(schema, dropout)
# LinearClassifier(output_dim, dropout)



class GenericLightningModule(pl.LightningModule):
    """
    All Zoobot models use the lightningmodule API and so share this structure
    super generic, just to outline the structure. nothing specific to dirichlet, gz, etc
    only assumes an encoder and a head
    """

    def __init__(
        self,
        *args,  # to be saved as hparams
        ):
        super().__init__()
        self.save_hyperparameters()  # saves all args by default
        self.setup_metrics()


    def setup_metrics(self):
        # these are ignored unless output dim = 2
        self.train_accuracy = Accuracy(task='binary')
        self.val_accuracy = Accuracy(task='binary')
        # self.log_on_step = False
        # self.log_on_step is useful for debugging, but slower - best when log_every_n_steps is fairly large


    def forward(self, x):
        x = self.encoder(x)
        return self.head(x)
    
    def make_step(self, batch, batch_idx, step_name):
        x, labels = batch
        predictions = self(x)  # by default, these are Dirichlet concentrations
        loss = self.calculate_and_log_loss(predictions, labels, step_name)      
        return {'loss': loss, 'predictions': predictions, 'labels': labels}

    def calculate_and_log_loss(self, predictions, labels, step_name):
        raise NotImplementedError('Must be subclassed')

    def configure_optimizers(self):
        raise NotImplementedError('Must be subclassed')

    def training_step(self, batch, batch_idx):
        return self.make_step(batch, batch_idx, step_name='train')

    def training_step_end(self, outputs):
        self.log_outputs(outputs, step_name='train')

    def validation_step(self, batch, batch_idx):
        return self.make_step(batch, batch_idx, step_name='validation')

    def validation_step_end(self, outputs):
        self.log_outputs(outputs, step_name='validation')

    def test_step(self, batch, batch_idx):
        return self.make_step(batch, batch_idx, step_name='test')

    def test_step_end(self, outputs):
         self.log_outputs(outputs, step_name='test')

    
    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        # https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#inference
        # this calls forward, while avoiding the need for e.g. model.eval(), torch.no_grad()
        # x, y = batch  # would be usual format, but here, batch does not include labels
        return self(batch)


class ZoobotTree(GenericLightningModule):
    """
    
    The Zoobot model. Train from scratch using :func:`zoobot.pytorch.training.train_with_pytorch_lightning.train_default_zoobot_from_scratch`.

    PyTorch LightningModule describing how to train the encoder and head (described below).
    Trains using Dirichlet loss. Labels should be num. volunteers giving each answer to each question. 

    See the code for exact training step, logging, etc - there's a lot of detail.

    Args:
        output_dim (int): Output dimension of model's head e.g. 34 for predicting a 34-answer decision tree.
        question_index_groups (List): Mapping of which label indices are part of the same question. See :ref:`training_on_vote_counts`.
        architecture_name (str, optional): Architecture to use. Passed to timm. Must be in timm.list_models(). Defaults to "efficientnet_b0".
        channels (int, optional): Num. input channels. Probably 3 or 1. Defaults to 1.
        use_imagenet_weights (bool, optional): Load weights pretrained on ImageNet (NOT galaxies!). Defaults to False.
        test_time_dropout (bool, optional): Apply dropout at test time, to pretend to be Bayesian. Defaults to True.
        timm_kwargs (dict, optional): passed to timm.create_model e.g. drop_path_rate=0.2 for effnet. Defaults to {}.
        learning_rate (float, optional): AdamW learning rate. Defaults to 1e-3.
    """

    # lightning only supports checkpoint loading / hparams which are not fancy classes
    # therefore, can't nicely wrap these arguments. So it goes.
    # override GenericLightningModule above, only this init
    def __init__(
        self,
        output_dim: int,
        question_index_groups: List,
        # encoder args
        architecture_name="efficientnet_b0",
        channels=1,
        use_imagenet_weights=False,
        test_time_dropout=True,
        timm_kwargs={},  # passed to timm.create_model e.g. drop_path_rate=0.2 for effnet
        # head args
        dropout_rate=0.2,
        learning_rate=1e-3,  # PyTorch default
        # optim args
        betas=(0.9, 0.999),  # PyTorch default
        weight_decay=0.01,  # AdamW PyTorch default
        scheduler_params={}  # no scheduler by default
        ):

        # now, finally, can pass only standard variables as hparams to save
        # will still need to actually use these variables later, this super init only saves them
        super().__init__(
            output_dim,
            question_index_groups,
            architecture_name,
            channels,
            timm_kwargs,
            test_time_dropout,
            dropout_rate,
            learning_rate,
            betas,
            weight_decay,
            scheduler_params
        )

        logging.info('Generic __init__ complete - moving to Zoobot __init__')

        # set attributes for learning rate, betas, used by self.configure_optimizers()
        # TODO refactor to optimizer params
        self.learning_rate = learning_rate
        self.betas = betas
        self.weight_decay = weight_decay
        self.scheduler_params = scheduler_params

        self.encoder = get_pytorch_encoder(
            architecture_name,
            channels,
            use_imagenet_weights=use_imagenet_weights,
            **timm_kwargs
        )
        # bit lazy assuming 224 input size
        self.encoder_dim = get_encoder_dim(self.encoder, input_size=224, channels=channels)
        # typically encoder_dim=1280 for effnetb0

        self.head = get_pytorch_dirichlet_head(
            self.encoder_dim,
            output_dim=output_dim,
            test_time_dropout=test_time_dropout,
            dropout_rate=dropout_rate
        )

        self.loss_func = get_dirichlet_loss_func(question_index_groups)

        logging.info('Zoobot __init__ complete')


    def calculate_and_log_loss(self, predictions, labels, step_name):
        # self.loss_func returns shape of (galaxy, question), mean to ()
        multiq_loss = self.loss_func(predictions, labels, sum_over_questions=False)
        # if hasattr(self, 'schema'):
        self.log_loss_per_question(multiq_loss, prefix=step_name)
        # sum over questions and take a per-device mean
        # for DDP strategy, batch size is constant (batches are not divided, data pool is divided)
        # so this will be the global per-example mean
        loss = torch.mean(torch.sum(multiq_loss, axis=1))
        return loss


    def configure_optimizers(self):
        # designed for training from scratch
        # parameters = list(self.head.parameters()) + list(self.encoder.parameters()) TODO should happen automatically?
        optimizer = torch.optim.AdamW(
            self.parameters(),
            lr=self.learning_rate,
            betas=self.betas,
            weight_decay=self.weight_decay
        )  
        if self.scheduler_params.get('name', None) == 'plateau':
            logging.info(f'Using Plateau scheduler with {self.scheduler_params}')
            # TODO could generalise this if needed
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer, 
                min_lr=1e-6,
                patience=self.scheduler_params.get('patience', 5)
            )
            return {'optimizer': optimizer, 'lr_scheduler': scheduler, 'monitor': 'validation/epoch_loss'}
        else:
            logging.info('No scheduler used')
            return optimizer  # no scheduler


    def log_outputs(self, outputs, step_name):
        self.log("{}/epoch_loss".format(step_name), outputs['loss'], on_epoch=True, on_step=False,prog_bar=True, logger=True, sync_dist=True)
        # if self.log_on_step:
        #     # seperate call to allow for different name, to allow for consistency with TF.keras auto-names
        #     self.log(
        #         "{}/step_loss".format(step_name), outputs['loss'], on_epoch=False, on_step=True, prog_bar=True, logger=True, sync_dist=True)
        if outputs['predictions'].shape[1] == 2:  # will only do for binary classifications
            # logging.info(predictions.shape, labels.shape)
            self.log(
                "{}_accuracy".format(step_name), self.train_accuracy(outputs['predictions'], torch.argmax(outputs['labels'], dim=1, keepdim=False)), prog_bar=True, sync_dist=True)


    def log_loss_per_question(self, multiq_loss, prefix):
        # log questions individually
        # TODO need schema attribute or similar to have access to question names, this will do for now
        for question_n in range(multiq_loss.shape[1]):
            self.log(f'{prefix}/epoch_questions/question_{question_n}_loss:0', torch.mean(multiq_loss[:, question_n]), on_epoch=True, on_step=False, sync_dist=True)


    

def get_dirichlet_loss_func(question_index_groups):
    # This just adds schema.question_index_groups as an arg to the usual (labels, preds) loss arg format
    # Would use lambda but multi-gpu doesn't support as lambda can't be pickled
    return partial(dirichlet_loss, question_index_groups=question_index_groups)


    # accept (labels, preds), return losses of shape (batch, question)
def dirichlet_loss(preds, labels, question_index_groups, sum_over_questions=False):
    # pytorch convention is preds, labels for loss func
    # my and sklearn convention is labels, preds for loss func

    # multiquestion_loss returns loss of shape (batch, question)
    # torch.sum(multiquestion_loss, axis=1) gives loss of shape (batch). Equiv. to non-log product of question likelihoods.
    multiq_loss = losses.calculate_multiquestion_loss(labels, preds, question_index_groups)
    if sum_over_questions:
        return torch.sum(multiq_loss, axis=1)
    else:
        return multiq_loss


def get_encoder_dim(encoder, input_size, channels):
    x = torch.randn(1, channels, input_size, input_size)  # batch size of 1
    return encoder(x).shape[-1]


def get_pytorch_encoder(
    architecture_name='efficientnet_b0',
    channels=1,
    use_imagenet_weights=False,
    **timm_kwargs
    ) -> nn.Module:
    """
    Create a trainable efficientnet model.
    First layers are galaxy-appropriate augmentation layers - see :meth:`zoobot.estimators.define_model.add_augmentation_layers`.
    Expects single channel image e.g. (300, 300, 1), likely with leading batch dimension.

    Optionally (by default) include the head (output layers) used for GZ DECaLS.
    Specifically, global average pooling followed by a dense layer suitable for predicting dirichlet parameters.
    See ``efficientnet_custom.custom_top_dirichlet``

    Args:
        output_dim (int): Dimension of head dense layer. No effect when include_top=False.
        input_size (int): Length of initial image e.g. 300 (asmeaned square)
        crop_size (int): Length to randomly crop image. See :meth:`zoobot.estimators.define_model.add_augmentation_layers`.
        resize_size (int): Length to resize image. See :meth:`zoobot.estimators.define_model.add_augmentation_layers`.
        weights_loc (str, optional): If str, load weights from efficientnet checkpoint at this location. Defaults to None.
        include_top (bool, optional): If True, include head used for GZ DECaLS: global pooling and dense layer. Defaults to True.
        expect_partial (bool, optional): If True, do not raise partial match error when loading weights (likely for optimizer state). Defaults to False.
        channels (int, default 1): Number of channels i.e. C in NHWC-dimension inputs. 

    Returns:
        torch.nn.Sequential: trainable efficientnet model including augmentations and optional head
    """
    # num_classes=0 gives pooled encoder
    # https://github.com/rwightman/pytorch-image-models/blob/main/timm/models/efficientnet.py
    
    # support older code that didn't specify effnet version
    if architecture_name == 'efficientnet':
        logging.warning('efficientnet variant not specified - please set architecture_name=efficientnet_b0 (or similar)')
        architecture_name = 'efficientnet_b0'
    return timm.create_model(architecture_name, in_chans=channels, num_classes=0, pretrained=use_imagenet_weights, **timm_kwargs)


def get_pytorch_dirichlet_head(encoder_dim: int, output_dim: int, test_time_dropout: bool, dropout_rate: float) -> torch.nn.Sequential:
    """
    Head to combine with encoder (above) when predicting Galaxy Zoo decision tree answers.
    Pytorch Sequential model.
    Predicts Dirichlet concentration parameters.
    
    Also used when finetuning on a new decision tree - see :class:`zoobot.pytorch.training.finetune.FinetuneableZoobotTree`.

    Args:
        encoder_dim (int): dimensions of preceding encoder i.e. the input size expected by this submodel.
        output_dim (int): output dimensions of this head e.g. 34 to predict 34 answers.
        test_time_dropout (bool): Use dropout at test time. 
        dropout_rate (float): P of dropout. See torch.nn.Dropout docs.

    Returns:
        torch.nn.Sequential: pytorch model expecting `encoder_dim` vector and predicting `output_dim` decision tree answers.
    """

    modules_to_use = []

    assert output_dim is not None
    # no AdaptiveAvgPool2d, encoder assumed to pool already  
    if test_time_dropout:
        logging.info('Using test-time dropout')
        dropout_layer = custom_layers.PermaDropout
    else:
        logging.info('Not using test-time dropout')
        dropout_layer = torch.nn.Dropout
    modules_to_use.append(dropout_layer(dropout_rate))
    # TODO could optionally add a bottleneck layer here
    modules_to_use.append(efficientnet_custom.custom_top_dirichlet(encoder_dim, output_dim))

    return nn.Sequential(*modules_to_use)
